import io
import logging
import pandas
from metta import metta_io as metta
import os


class BuilderBase(object):
    def __init__(self, db_config, matrix_directory, engine, replace=True):
        self.db_config = db_config
        self.matrix_directory = matrix_directory
        self.engine = engine
        self.replace = replace

    def validate(self):
        for expected_db_config_val in [
            'features_schema_name',
            'sparse_state_table_name',
            'labels_schema_name',
            'labels_table_name'
        ]:
            if expected_db_config_val not in self.db_config:
                raise ValueError('{} needed in db_config'.format(expected_db_config_val))

    def build_all_matrices(self, build_tasks):
        logging.info('Building %s matrices', len(build_tasks.keys()))
        for matrix_uuid, task_arguments in build_tasks.items():
            self.build_matrix(**task_arguments)

    def _outer_join_query(
        self,
        right_table_name,
        right_column_selections,
        entity_date_table_name,
        additional_conditions=''
    ):
        """ Given a (features or labels) table, a list of times, columns to
        select, and (optionally) a set of join conditions, perform an outer
        join to the entity date table.

        :param right_table_name: the name of the right (feature/label) table
        :param right_column_selections: formatted text for the columns to select
        :param entity_date_table_name: name of table containing all valid entity ids and dates
        :param additional_conditions: formatted text for additional join
                                      conditions
        :type right_table_name: str
        :type right_column_selections: str
        :type entity_date_table_name: str
        :type additional_conditions: str

        :return: postgresql query for the outer join to the entity-dates table
        :rtype: str
        """

        # put everything into the query
        query = """
            SELECT ed.entity_id,
                   ed.as_of_date{columns}
            FROM {entity_date_table_name} ed
            LEFT OUTER JOIN {right_table} r
            ON ed.entity_id = r.entity_id AND
               ed.as_of_date = r.as_of_date
               {more}
            ORDER BY ed.entity_id,
                     ed.as_of_date
        """.format(
            columns=''.join(right_column_selections),
            feature_schema=self.db_config['features_schema_name'],
            entity_date_table_name=entity_date_table_name,
            right_table=right_table_name,
            more=additional_conditions
        )
        return(query)

    def make_entity_date_table(
        self,
        as_of_times,
        label_name,
        label_type,
        state,
        matrix_type,
        matrix_uuid,
        label_timespan
    ):
        """ Make a table containing the entity_ids and as_of_dates required for
        the current matrix.

        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param state: the entity state to be used in the matrix
        :param matrix_type: the type (train/test) of matrix
        :param matrix_uuid: a unique id for the matrix
        :param label_timespan: the time timespan that labels in matrix will include
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type state: str
        :type matrix_type: str
        :type matrix_uuid: str
        :type label_timespan: str

        :return: table name
        :rtype: str
        """

        as_of_time_strings = [str(as_of_time) for as_of_time in as_of_times]
        if matrix_type == 'train':
            indices_query = self._all_labeled_entity_dates_query(
                as_of_time_strings=as_of_time_strings,
                state=state,
                label_name=label_name,
                label_type=label_type,
                label_timespan=label_timespan
            )
        elif matrix_type == 'test':
            indices_query = self._all_valid_entity_dates_query(
                as_of_time_strings=as_of_time_strings,
                state=state
            )
        else:
            raise ValueError('Unknown matrix type passed: {}'.format(matrix_type))

        table_name = '_'.join([matrix_uuid, 'matrix_entity_date'])
        query = """
            DROP TABLE IF EXISTS {features_schema_name}."{table_name}";
            CREATE TABLE {features_schema_name}."{table_name}"
            AS ({index_query})
        """.format(
            features_schema_name=self.db_config['features_schema_name'],
            table_name=table_name,
            index_query=indices_query
        )
        logging.info('Creating matrix-specific entity-date table for matrix %s with query %s', matrix_uuid, query)
        self.engine.execute(query)

        return table_name

    def _all_labeled_entity_dates_query(
        self,
        as_of_time_strings,
        state,
        label_name,
        label_type,
        label_timespan
    ):
        query = """
            SELECT entity_id, as_of_date
            FROM {states_table}
            JOIN {labels_schema_name}.{labels_table_name} using (entity_id, as_of_date)
            WHERE {state_string}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{times}::timestamp[])))
            AND label_name = '{l_name}'
            AND label_type = '{l_type}'
            AND label_timespan = '{timespan}'
            AND label is not null
            ORDER BY entity_id, as_of_date
        """.format(
            states_table=self.db_config['sparse_state_table_name'],
            state_string=state,
            labels_schema_name=self.db_config['labels_schema_name'],
            labels_table_name=self.db_config['labels_table_name'],
            l_name=label_name,
            l_type=label_type,
            timespan=label_timespan,
            times=as_of_time_strings
        )

        return query

    def _all_valid_entity_dates_query(self, state, as_of_time_strings):
        query = """
            SELECT entity_id, as_of_date
            FROM {states_table}
            WHERE {state_string}
            AND as_of_date IN (SELECT (UNNEST (ARRAY{times}::timestamp[])))
            ORDER BY entity_id, as_of_date
        """.format(
            states_table=self.db_config['sparse_state_table_name'],
            state_string=state,
            times=as_of_time_strings
        )
        return query


class CSVBuilder(BuilderBase):
    def build_matrix(
        self,
        as_of_times,
        label_name,
        label_type,
        feature_dictionary,
        matrix_directory,
        matrix_metadata,
        matrix_uuid,
        matrix_type
    ):
        """ Write a design matrix to disk with the specified paramters.

        :param as_of_times: datetimes to be included in the matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param feature_dictionary: a dictionary of feature tables and features
                                   to be included in the matrix
        :param matrix_directory: the directory in which to store the matrix
        :param matrix_metadata: a dictionary of metadata about the matrix
        :param matrix_uuid: a unique id for the matrix
        :param matrix_type: the type (train/test) of matrix
        :type as_of_times: list
        :type label_name: str
        :type label_type: str
        :type feature_dictionary: dict
        :type matrix_directory: str
        :type matrix_metadata: dict
        :type matrix_uuid: str
        :type matrix_type: str

        :return: none
        :rtype: none
        """
        matrix_filename = os.path.join(
            matrix_directory,
            '{}.csv'.format(matrix_uuid)
        )
        if not self.replace and os.path.exists(matrix_filename):
            logging.info('Skipping %s because matrix already exists', matrix_filename)
            return

        logging.info('Creating matrix %s > %s', matrix_metadata['matrix_id'], matrix_filename)
        # make the entity time table and query the labels and features tables
        logging.info('Making entity date table for matrix %s', matrix_uuid)
        entity_date_table_name = self.make_entity_date_table(
            as_of_times,
            label_name,
            label_type,
            matrix_metadata['state'],
            matrix_type,
            matrix_uuid,
            matrix_metadata['label_timespan']
        )
        logging.info('Extracting feature group data from database into file for matrix %s', matrix_uuid)
        features_csv_names = self.write_features_data(
            as_of_times,
            feature_dictionary,
            entity_date_table_name,
            matrix_uuid
        )
        try:
            logging.info('Extracting label data frmo database into file for matrix %s', matrix_uuid)
            labels_csv_name = self.write_labels_data(
                label_name,
                label_type,
                entity_date_table_name,
                matrix_uuid,
                matrix_metadata['label_timespan']
            )
            features_csv_names.insert(0, labels_csv_name)

            # stitch together the csvs
            logging.info('Merging feature files for matrix %s', matrix_uuid)
            output = self.merge_feature_csvs(
                features_csv_names,
                matrix_directory,
                matrix_uuid
            )
        finally:
            # clean up files and database before finishing
            for csv_name in features_csv_names:
                self.remove_file(csv_name)
        try:
            # store the matrix
            logging.info('Archiving matrix %s with metta', matrix_uuid)
            metta.archive_matrix(
                matrix_config=matrix_metadata,
                df_matrix=output,
                overwrite=True,
                directory=self.matrix_directory,
                format='csv'
            )
        finally:
            if isinstance(output, str):
                os.remove(output)

    def write_labels_data(
        self,
        label_name,
        label_type,
        entity_date_table_name,
        matrix_uuid,
        label_timespan
    ):
        """ Query the labels table and write the data to disk in csv format.

        :param as_of_times: the times to be used for the current matrix
        :param label_name: name of the label to be used
        :param label_type: the type of label to be used
        :param entity_date_table_name: the name of the entity date table
        :param matrix_uuid: a unique id for the matrix
        :param label_timespan: the time timespan that labels in matrix will include
        :type label_name: str
        :type label_type: str
        :type entity_date_table_name: str
        :type matrix_uuid: str
        :type label_timespan: str

        :return: name of csv containing labels
        :rtype: str
        """
        csv_name = os.path.join(
            self.matrix_directory,
            '{}-{}.csv'.format(matrix_uuid, self.db_config['labels_table_name'])
        )
        labels_query = self._outer_join_query(
            right_table_name='{schema}.{table}'.format(
                schema=self.db_config['labels_schema_name'],
                table=self.db_config['labels_table_name']
            ),
            entity_date_table_name='"{schema}"."{table}"'.format(
                schema=self.db_config['features_schema_name'],
                table=entity_date_table_name
            ),
            right_column_selections=', r.label as {}'.format(label_name),
            additional_conditions='''AND
                r.label_name = '{name}' AND
                r.label_type = '{type}' AND
                r.label_timespan = '{timespan}'
            '''.format(
                name=label_name,
                type=label_type,
                timespan=label_timespan
            )
        )

        self.write_to_csv(labels_query, csv_name)
        return(csv_name)

    def write_features_data(
        self,
        as_of_times,
        feature_dictionary,
        entity_date_table_name,
        matrix_uuid
    ):
        """ Loop over tables in features schema, writing the data from each to a
        csv. Return the full list of feature csv names and the list of all
        features.

        :param as_of_times: the times to be included in the matrix
        :param feature_dictionary: a dictionary of feature tables and features
            to be included in the matrix
        :param entity_date_table_name: the name of the entity date table
            for the matrix
        :param matrix_uuid: a human-readable id for the matrix
        :type as_of_times: list
        :type feature_dictionary: dict
        :type entity_date_table_name: str
        :type matrix_uuid: str

        :return: list of csvs containing feature data
        :rtype: tuple
        """
        # iterate! for each table, make query, write csv, save feature & file names
        features_csv_names = []
        for feature_table_name, feature_names in feature_dictionary.items():
            logging.info('Retrieving feature data from %s', feature_table_name)
            csv_name = os.path.join(
                self.matrix_directory,
                '{}-{}.csv'.format(matrix_uuid, feature_table_name)
            )
            features_query = self._outer_join_query(
                right_table_name='{schema}.{table}'.format(
                    schema=self.db_config['features_schema_name'],
                    table=feature_table_name
                ),
                entity_date_table_name='{schema}."{table}"'.format(
                    schema=self.db_config['features_schema_name'],
                    table=entity_date_table_name
                ),
                # collate imputation shouldn't leave any nulls and we double-check
                # the imputed table in FeatureGenerator.create_all_tables() but as
                # a final check, raise a divide by zero error on export if the
                # database encounters any during the outer join
                right_column_selections=[', "{0}"'.format(fn) for fn in feature_names]
            )
            self.write_to_csv(features_query, csv_name)
            features_csv_names.append(csv_name)

        return(features_csv_names)

    def write_to_csv(self, query_string, file_name, header='HEADER'):
        """ Given a query, write the requested data to csv.

        :param query_string: query to send
        :param file_name: name to save the file as
        :header: text to include in query indicating if a header should be saved
                 in output
        :type query_string: str
        :type file_name: str
        :type header: str

        :return: none
        :rtype: none
        """
        matrix_csv = self.open_fh_for_writing(file_name)
        logging.debug('Copying to CSV query %s', query_string)
        try:
            copy_sql = 'COPY ({query}) TO STDOUT WITH CSV {head}'.format(
                query=query_string,
                head=header
            )
            conn = self.engine.raw_connection()
            cur = conn.cursor()
            cur.copy_expert(copy_sql, matrix_csv)
        finally:
            self.close_filehandle(file_name)


class HighMemoryCSVBuilder(CSVBuilder):
    def __init__(self, *args, **kwargs):
        super(HighMemoryCSVBuilder, self).__init__(*args, **kwargs)
        self.filehandles = {}

    def open_fh_for_writing(self, filename):
        self.filehandles[filename] = io.StringIO()
        return self.filehandles[filename]

    def open_fh_for_reading(self, filename):
        self.filehandles[filename].seek(0)
        return self.filehandles[filename]

    def close_filehandles(self):
        pass

    def close_filehandle(self, filename):
        pass

    def remove_file(self, filename):
        del self.filehandles[filename]

    def merge_feature_csvs(self, source_filenames, matrix_directory, matrix_uuid):
        """Horizontally merge a list of feature CSVs
        Assumptions:
        - The first and second columns of each CSV are
          the entity_id and date
        - That the CSVs have the same list of entity_id/date combinations
          in the same order.
        - The first CSV is expected to be labels, and only have
          entity_id, date, and label.
        - All other CSVs do not have any labels (all non entity_id/date columns
          will be treated as features)
        - The label will be in the *last* column of the merged CSV

        :param source_filenames: the filenames of each feature csv
        :param out_filename: the desired filename of the merged csv
        :type source_filenames: list
        :type out_filename: str

        :return: none
        :rtype: none

        :raises: ValueError if the first two columns in every CSV don't match
        """

        source_filehandles = [self.open_fh_for_reading(fname) for fname in source_filenames]
        dataframes = []
        for i, filehandle in enumerate(source_filehandles):
            df = pandas.read_csv(filehandle)
            df.set_index(['entity_id', 'as_of_date'], inplace=True)
            dataframes.append(df)

            # check for any nulls. the labels, understood to be the first file,
            # can have nulls but no features should. therefore, skip the first dataframe
            if i > 0:
                columns_with_nulls = [
                    column
                    for column in df.columns
                    if df[column].isnull().values.any()
                ]
                if len(columns_with_nulls) > 0:
                    raise ValueError(
                        "Imputation failed for the following features: %s" %
                        columns_with_nulls
                    )
            i += 1

        big_df = dataframes[1].join(dataframes[2:] + [dataframes[0]])
        return big_df



from architect.feature_generators import FeatureGenerator
from architect.feature_dictionary_creator import FeatureDictionaryCreator
from architect.feature_group_creator import FeatureGroupCreator
from architect.feature_group_mixer import FeatureGroupMixer



import logging

from architect.database_reflection import table_has_data
from architect.validations import (
    table_should_have_data,
    column_should_be_intlike,
    column_should_be_booleanlike,
    column_should_be_timelike,
    column_should_be_stringlike,
)


class StateFilter(object):

    def __init__(self, sparse_state_table, filter_logic):
        self.sparse_state_table = sparse_state_table
        self.filter_logic = filter_logic

    def join_sql(self, join_table, join_date):
        return '''
            join {state_table} on (
                {state_table}.entity_id = {join_table}.entity_id and
                {state_table}.as_of_date = '{join_date}'::timestamp and
                ({state_filter_logic})
            )
        '''.format(
            state_table=self.sparse_state_table,
            state_filter_logic=self.filter_logic,
            join_date=join_date,
            join_table=join_table
        )


DEFAULT_ACTIVE_STATE = 'active'


class StateTableGenerator(object):
    """Create a table containing the state of entities on different dates

    Throughout the class we will refer to two types of state tables:
        'dense' and 'sparse'.

    dense: A table containing entities and *time ranges* for different states

        An example of this would be a building permits table, with
        time ranges that buildings are permitted.

        The expected format would be entity id/state/start/end

    sparse: A table containing entities, dates, and boolean values for
        different states.

        These types of tables rarely exist in source data, but are useful
        internally in pipelines to express state logic in SQL

        For instance: '(in_jail OR booked) AND in_mental_health_facility'

        The output format is entity id/date/state1/state3/state3...

    The main interface of StateTableGenerator objects is the
    `generate_sparse_table` method, which produces the latter
    'sparse'-style table.

    Args:
        db_engine (sqlalchemy.engine)
        experiment_hash (string) unique identifier for the experiment
        events_table (string, optional) name of SQL table containing
            outcome events for entities
        dense_state_table (string, optional) name of SQL table containing
            entities and state time ranges

    """
    def __init__(
        self,
        db_engine,
        experiment_hash,
        events_table=None,
        dense_state_table=None
    ):
        self.db_engine = db_engine
        self.experiment_hash = experiment_hash
        self.dense_state_table = dense_state_table
        self.events_table = events_table

    def validate(self):
        if not self.dense_state_table and not self.events_table:
            raise ValueError('Need to specify either dense_state_table or events_table')
        if self.events_table:
            self._event_validations()
        if self.dense_state_table:
            self._dense_state_validations()

    def _dense_state_validations(self):
        table_should_have_data(self.dense_state_table, self.db_engine)

        column_should_be_intlike(self.dense_state_table, 'entity_id', self.db_engine)
        column_should_be_stringlike(self.dense_state_table, 'state', self.db_engine)
        column_should_be_timelike(self.dense_state_table, 'start_time', self.db_engine)
        column_should_be_timelike(self.dense_state_table, 'end_time', self.db_engine)

    def _event_validations(self):
        table_should_have_data(self.events_table, self.db_engine)
        column_should_be_intlike(self.events_table, 'entity_id', self.db_engine)
        column_should_be_timelike(self.events_table, 'outcome_date', self.db_engine)
        column_should_be_booleanlike(self.events_table, 'outcome', self.db_engine)

    @property
    def sparse_table_name(self):
        return 'tmp_sparse_states_{}'.format(self.experiment_hash)

    @property
    def sparse_table_query_func(self):
        if self.dense_state_table:
            logging.info('Dense state table passed to StateTableGenerator, so computing sparse table using it')
            return self._sparse_table_query_from_dense
        else:
            logging.info('Dense state table not passed to StateTableGenerator, so computing sparse table using events table')
            return self._sparse_table_query_from_events

    def _all_known_states(self, dense_state_table):
        all_states = [
            row[0] for row in
            self.db_engine.execute('''
                select distinct(state) from {} order by state
            '''.format(dense_state_table))
        ]
        logging.info('Distinct states found: %s', all_states)
        return all_states

    def _sparse_table_query_from_dense(self, as_of_dates):
        """A query to convert a dense-style state table to a 'sparse'-style
        table containing a specific set of dates.

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of

        Returns: (string) A query to produce a sparse states table
        """
        state_columns = [
            "bool_or(state = '{desired_state}') as {desired_state}"
            .format(desired_state=state)
            for state in self._all_known_states(self.dense_state_table)
        ]
        if not state_columns:
            raise ValueError("Unable to identify states from table",
                             self.dense_state_table)
        query = '''
            create table {sparse_state_table} as (
            select d.entity_id, a.as_of_date::timestamp, {state_column_string}
                from {dense_state_table} d
                join (select unnest(ARRAY{as_of_dates}) as as_of_date) a
                on (
                    d.start_time <= a.as_of_date::timestamp and
                    d.end_time > a.as_of_date::timestamp
                )
                group by d.entity_id, a.as_of_date
            )
        '''.format(
            sparse_state_table=self.sparse_table_name,
            dense_state_table=self.dense_state_table,
            as_of_dates=[date.isoformat() for date in as_of_dates],
            state_column_string=', '.join(state_columns)
        )
        logging.debug('Assembled sparse state table query: %s', query)
        return query

    def _sparse_table_query_from_events(self, as_of_dates):
        """A query to convert an events table to a 'sparse'-style
        table containing a specific set of dates.

        This will include all entities for all given dates

        Args:
        as_of_dates (list of datetime.date): Dates to calculate entity states as of

        Returns: (string) A query to produce a sparse states table
        """

        query = '''
            create table {sparse_state_table} as (
            select e.entity_id, a.as_of_date::timestamp, true {active_state}
                from {events_table} e
                cross join (select unnest(ARRAY{as_of_dates}) as as_of_date) a
                group by e.entity_id, a.as_of_date
            )
        '''.format(
            sparse_state_table=self.sparse_table_name,
            events_table=self.events_table,
            as_of_dates=[date.isoformat() for date in as_of_dates],
            active_state=DEFAULT_ACTIVE_STATE
        )
        logging.debug('Assembled sparse state table query: %s', query)
        return query

    def generate_sparse_table(self, as_of_dates):
        """Convert the object's input table (either dense states or events)
        into a sparse states table for the given as_of_dates

        Args:
            as_of_dates (list of datetime.dates) Dates to include in the sparse
                state table
        """
        logging.debug('Generating sparse table using as_of_dates: %s', as_of_dates)
        self._generate_sparse_table(self.sparse_table_query_func(as_of_dates))
        if not table_has_data(self.sparse_table_name, self.db_engine):
            raise ValueError(
                "No entities in table '{input_table}' define time ranges "
                "that encompass any of experiment's \"as-of-dates\":\n\n"
                "\t{as_of_dates}\n\n"
                "Please check temporal and/or state configurations."
                .format(
                    input_table=(self.dense_state_table or self.events_table),
                    as_of_dates=', '.join(str(as_of_date) for as_of_date in (
                        as_of_dates if len(as_of_dates) <= 5 else as_of_dates[:5] + ['…']
                    )),
                )
            )

    def _generate_sparse_table(self, generate_query):
        """Generate and index a sparse table from a given query

        Args:
            generate_query (string) A full query to generate a sparse table
        """
        self.db_engine.execute(generate_query)
        logging.info('Sparse state table generated')
        self.db_engine.execute(
            'create index on {} (entity_id, as_of_date)'
            .format(self.sparse_table_name)
        )
        logging.info('Indices created on entity_id and as_of_date for sparse state table')

    def clean_up(self):
        self.db_engine.execute(
            'drop table if exists {}'.format(self.sparse_table_name)
        )



from datetime import datetime
from dateutil.relativedelta import relativedelta
import functools
import operator
import warnings
import re


def convert_str_to_relativedelta(delta_string):
    """ Given a string in a postgres interval format (e.g., '1 month'),
    convert it to a dateutil.relativedelta.relativedelta.

    Assumptions:
    - the label_timespan string is in the format 'value unit', where
      value is an int and unit is one of year(s), month(s), day(s),
      week(s), hour(s), minute(s), second(s), microsecond(s).

    :param delta_string: the time interval to convert
    :type delta_string: str

    :return: the time interval as a relativedelta
    :rtype: dateutil.relativedelta.relativedelta

    :raises: ValueError if the delta_string is not in the expected format
    """
    units, value = parse_delta_string(delta_string)

    if units in ['year', 'years', 'y', 'Y']:
        delta = relativedelta(years = value)
    elif units in ['month', 'months']:
        delta = relativedelta(months = value)
    elif units in ['day', 'days', 'd', 'D']:
        delta = relativedelta(days = value)
    elif units in ['week', 'weeks', 'w', 'W']:
        delta = relativedelta(weeks = value)
    elif units in ['hour', 'hours', 'h', 'H']:
        delta = relativedelta(hours = value)
    elif units in ['minute', 'minutes', 'm', 'M']:
        delta = relativedelta(minutes = value)
        if units in ['m', 'M']:
            warnings.warn(
                'Time delta units "{}" converted to minutes.'.format(units),
                RuntimeWarning
            )
    elif units in ['second', 'seconds', 's', 'S']:
        delta = relativedelta(seconds = value)
    elif units == 'microsecond' or units == 'microseconds':
        delta = relativedelta(microseconds = value)
    else:
        raise ValueError(
            'Could not handle units. Units: {} Value: {}'.format(units, value)
        )

    return(delta)

def parse_delta_string(delta_string):
    if len(delta_string.split(' ')) == 2:
        units = delta_string.split(' ')[1]
        try:
            value = int(delta_string.split(' ')[0])
        except:
            raise ValueError('''
                Could not parse value from time delta string: {}
            '''.format(delta_string))
    else:
        delta_parts = re.split('([a-zA-Z]*)', delta_string)
        units = delta_parts[1]
        try:
            value = int(delta_parts[0])
        except:
            raise ValueError('''
                Could not parse value from time delta string: {}
            '''.format(delta_string))
    return(units, value)


def feature_list(feature_dictionary):
    """Convert a feature dictionary to a sorted list.

    Args: feature_dictionary (dict)

    Returns: sorted list of feature names
    """
    return sorted(functools.reduce(
        operator.concat,
        [feature_dictionary[key] for key in feature_dictionary.keys()]
    ))


def str_in_sql(values):
    return ','.join(map(lambda x: "'{}'".format(x), values))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('LICENSE') as license_file:
    license = license_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

with open('requirements_dev.txt') as dev_requirements_file:
    test_requirements = requirements + [
        line for line in dev_requirements_file.readlines()
        if '-r requirements' not in line
    ]


setup(
    name='matrix-architect',
    version='1.0.1',
    description="Plan, design, and build train and test matrices",
    long_description=readme,
    author="Center for Data Science and Public Policy",
    author_email='datascifellows@gmail.com',
    url='https://github.com/dssg/architect',
    packages=[
        'architect',
    ],
    include_package_data=True,
    install_requires=requirements,
    license=license,
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



import logging
from architect.validations import table_should_have_data,\
    column_should_be_intlike,\
    column_should_be_booleanlike,\
    column_should_be_timelike


class BinaryLabelGenerator(object):
    def __init__(self, events_table, db_engine):
        self.events_table = events_table
        self.db_engine = db_engine

    def validate(self):
        table_should_have_data(self.events_table, self.db_engine)
        column_should_be_intlike(self.events_table, 'entity_id', self.db_engine)
        column_should_be_timelike(self.events_table, 'outcome_date', self.db_engine)
        column_should_be_booleanlike(self.events_table, 'outcome', self.db_engine)

    def generate(
        self,
        start_date,
        label_timespan,
        labels_table,
    ):
        query = """insert into {labels_table} (
            select
                {events_table}.entity_id,
                '{start_date}'::date as as_of_date,
                '{label_timespan}'::interval as label_timespan,
                'outcome' as label_name,
                'binary' as label_type,
                bool_or(outcome::bool)::int as label
            from {events_table}
            where '{start_date}' <= outcome_date
            and outcome_date < '{start_date}'::timestamp + interval '{label_timespan}'
            group by 1, 2, 3, 4, 5
        )""".format(
            events_table=self.events_table,
            labels_table=labels_table,
            start_date=start_date,
            label_timespan=label_timespan,
        )
        logging.debug('Running label generation query: %s', query)
        self.db_engine.execute(query)
        return labels_table

    def _create_labels_table(self, labels_table_name):
        self.db_engine.execute(
            'drop table if exists {}'.format(labels_table_name)
        )
        self.db_engine.execute('''
            create table {} (
            entity_id int,
            as_of_date date,
            label_timespan interval,
            label_name varchar(30),
            label_type varchar(30),
            label int
        )'''.format(labels_table_name))

    def generate_all_labels(
        self,
        labels_table,
        as_of_dates,
        label_timespans,
    ):
        self._create_labels_table(labels_table)
        logging.info('Creating labels for %s as of dates and %s label timespans',
                     len(as_of_dates),
                     len(label_timespans))
        for as_of_date in as_of_dates:
            for label_timespan in label_timespans:
                logging.info('Generating labels for as of date %s and label timespan %s', as_of_date, label_timespan)
                self.generate(
                    start_date=as_of_date,
                    label_timespan=label_timespan,
                    labels_table=labels_table,
                )
        nrows = [
            row[0] for row in
            self.db_engine.execute('select count(*) from {}'.format(labels_table))
        ][0]
        if nrows == 0:
            logging.warning('Done creating labels, but no rows in labels table!')
        else:
            logging.info('Rows in labels table: %s', nrows)
